{-# LANGUAGE ImportQualifiedPost #-}
{-# LANGUAGE InstanceSigs #-}

-- | 
-- Copyright: (c) Luca S. Jaekel
-- License: AGPL3
--
-- Compressed Instructions are useful for brainfuck optimization, since they enable easier pattern-matching on common program patterns.

module Language.Brainfuck.Instruction.Compressed ( CompressedInstruction(..), compress, uncompress ) where

import Data.Bifunctor ( Bifunctor(first) )
import Data.Vector ( Vector )
import Data.Word ( Word8 )
import Numeric.Natural ( Natural )
import Test.QuickCheck.Gen (Gen)
import Test.QuickCheck.Arbitrary (Arbitrary(arbitrary))
import Test.QuickCheck.Instances.Natural ()

import Data.Vector qualified as Vector

import Test.QuickCheck.Gen qualified as Gen

import Language.Brainfuck.Instruction ( Instruction )
import Language.Brainfuck.Instruction qualified as Instruction 

-- | Represents a single dense instruction, repeatable instructions have a effect count associated with them. They should not occur next to each other in data structures when they could be merged in any way.

data CompressedInstruction
  = Add Word8
  | Subtract Word8
  -- add and sub wrap when doing more than 255 because Brainfuck operates on u8-cells
  | MoveRight Natural
  | MoveLeft Natural
  -- disallows moving a negative amount of steps in a direction
  | ReadByte
  | PutByte
  | Loop (Vector CompressedInstruction)
  deriving (Show, Eq)

-- >>> fromInteger 300 :: Word8
-- 44


instance Arbitrary CompressedInstruction where
  arbitrary :: Gen CompressedInstruction
  arbitrary = Gen.oneof
    [ Add <$> arbitrary
    , Subtract <$> arbitrary
    , MoveRight <$> arbitrary
    , MoveLeft <$> arbitrary
    , pure ReadByte
    , pure PutByte
    , Loop <$> resize (`div` 8) arbitrary
    ]
    where
      resize f x = Gen.sized $ \ s -> Gen.resize (f s) x

-- | Will count reoccuring instructions and associate the count with the repeatable instruction. This should reduce the size of the instruction vector for any useful program. 
--
-- >>> compress . Vector.fromList $ replicate 300 Instruction.Increment
-- [Add 44]
--
-- >>> compress . Vector.fromList $ replicate 300 Instruction.MoveRight
-- [MoveRight 300]

compress :: Vector Instruction -> Vector CompressedInstruction
compress instructions = Vector.fromList (go instructions)
  where
    go program = case Vector.uncons program of
      Nothing -> []
      Just (x, rest) -> let

        withRepeatCount f = let (count, rest') = spanLength x rest
          in (f . fromIntegral . succ) count : go rest'
          
        in case x of
          Instruction.Increment  -> withRepeatCount Add
          Instruction.Decrement  -> withRepeatCount Subtract
          Instruction.MoveRight  -> withRepeatCount MoveRight
          Instruction.MoveLeft   -> withRepeatCount MoveLeft
          Instruction.ReadByte   -> ReadByte : go rest
          Instruction.PutByte    -> PutByte : go rest
          Instruction.Loop inner -> Loop (compress inner) : go rest

    spanLength :: Eq a => a -> Vector a -> (Int, Vector a)
    spanLength x xs = first Vector.length $ Vector.span (== x) xs

-- | Uncompress is not necessarily the exact inverse of compress, but it is a brainfuck program semantic-preserving function.
--
-- >>> compress . Vector.fromList $ replicate 300 Instruction.Increment
-- [Add 44]
--
-- >>> let source = Vector.replicate 300 Instruction.Increment in source == (uncompress . compress) source
-- False

uncompress :: Vector CompressedInstruction -> Vector Instruction
uncompress = Vector.concatMap uncompressSingle
  where
    uncompressSingle instruction = let
      repeated x times = Vector.replicate (fromIntegral times) x

      in case instruction of
        Add times       -> repeated Instruction.Increment times
        Subtract times  -> repeated Instruction.Decrement times
        MoveRight times -> repeated Instruction.MoveRight times
        MoveLeft  times -> repeated Instruction.MoveLeft  times
        PutByte         -> Vector.singleton Instruction.PutByte
        ReadByte        -> Vector.singleton Instruction.ReadByte
        Loop      body  -> Vector.singleton $ Instruction.Loop (uncompress body)

